<?php
/**

 *
 * LICENSE
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://framework.zend.com/license/new-bsd
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@zend.com so we can send you a copy immediately.
 *
 * @category   Zend
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

/**
 * Class for sending an email.
 *
 * @category   Zend
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 */
class we_mail_znd extends we_mail_mimeMessage{
	const PROHIBITTED_MAIL_HEADERS = ['to', 'cc', 'bcc', 'from', 'subject', 'reply-to', 'return-path', 'date', 'message-id',];

	/*	 * #@+
	 * @access protected
	 */
	/**
	 * @var we_mail_TransportAbstract
	 * @static
	 */
	protected static $defaultTransport = null;

	/**
	 * @var array
	 * @static
	 */
	protected static $defaultFrom;

	/**
	 * @var array
	 * @static
	 */
	protected static $defaultReplyTo;

	/**
	 * Mail character set
	 * @var string
	 */
	protected $_charset = 'iso-8859-1';

	/**
	 * Mail headers
	 * @var array
	 */
	protected $_headers = [];

	/**
	 * Encoding of Mail headers
	 * @var string
	 */
	protected $_headerEncoding = we_mail_mime::ENCODING_QUOTEDPRINTABLE;

	/**
	 * From: address
	 * @var string
	 */
	protected $_from = null;

	/**
	 * To: addresses
	 * @var array
	 */
	protected $_to = [];

	/**
	 * Array of all recipients
	 * @var array
	 */
	protected $_recipients = [];

	/**
	 * Reply-To header
	 * @var string
	 */
	protected $_replyTo = null;

	/**
	 * Return-Path header
	 * @var string
	 */
	protected $_returnPath = null;

	/**
	 * Subject: header
	 * @var string
	 */
	protected $_subject = null;

	/**
	 * Date: header
	 * @var string
	 */
	protected $_date = null;

	/**
	 * Message-ID: header
	 * @var string
	 */
	protected $_messageId = null;

	/**
	 * text/plain MIME part
	 * @var false|we_mail_mimePart
	 */
	protected $_bodyText = false;

	/**
	 * text/html MIME part
	 * @var false|we_mail_mimePart
	 */
	protected $_bodyHtml = false;

	/**
	 * MIME boundary string
	 * @var string
	 */
	protected $_mimeBoundary = null;

	/**
	 * Content type of the message
	 * @var string
	 */
	protected $_type = null;

	/*	 * #@- */

	/**
	 * Flag: whether or not email has attachments
	 * @var boolean
	 */
	public $hasAttachments = false;

	/**
	 * Sets the default mail transport for all following uses of
	 * Mail::send();
	 *
	 * @todo Allow passing a string to indicate the transport to load
	 * @todo Allow passing in optional options for the transport to load
	 * @param  we_mail_TransportAbstract $transport
	 */
	public static function setDefaultTransport(we_mail_TransportAbstract $transport){
		self::$defaultTransport = $transport;
	}

	/**
	 * Gets the default mail transport for all following uses of
	 * unittests
	 *
	 * @todo Allow passing a string to indicate the transport to load
	 * @todo Allow passing in optional options for the transport to load
	 */
	public static function getDefaultTransport(){
		return self::$defaultTransport;
	}

	/**
	 * Clear the default transport property
	 */
	public static function clearDefaultTransport(){
		self::$defaultTransport = null;
	}

	/**
	 * Public constructor
	 *
	 * @param  string $charset
	 * @return void
	 */
	public function __construct($charset = null){
		if($charset != null){
			$this->_charset = $charset;
		}
	}

	/**
	 * Return charset string
	 *
	 * @return string
	 */
	public function getCharset(){
		return $this->_charset;
	}

	/**
	 * Set content type
	 *
	 * Should only be used for manually setting multipart content types.
	 *
	 * @param  string $type Content type
	 * @return we_mail_znd Implements fluent interface
	 * @throws we_mail_exception for types not supported by Mime
	 */
	public function setType($type){
		$allowed = [
			we_mail_mime::MULTIPART_ALTERNATIVE,
			we_mail_mime::MULTIPART_MIXED,
			we_mail_mime::MULTIPART_RELATED,
		];
		if(!in_array($type, $allowed)){
			/**
			 * @see Mail_Exception
			 */
			throw new we_mail_exception('Invalid content type "' . $type . '"');
		}

		$this->_type = $type;
		return $this;
	}

	/**
	 * Get content type of the message
	 *
	 * @return string
	 */
	public function getType(){
		return $this->_type;
	}

	/**
	 * Set an arbitrary mime boundary for the message
	 *
	 * If not set, Mime will generate one.
	 *
	 * @param  string    $boundary
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setMimeBoundary($boundary){
		$this->_mimeBoundary = $boundary;

		return $this;
	}

	/**
	 * Return the boundary string used for the message
	 *
	 * @return string
	 */
	public function getMimeBoundary(){
		return $this->_mimeBoundary;
	}

	/**
	 * Return encoding of mail headers
	 *
	 * @deprecated use {@link getHeaderEncoding()} instead
	 * @return string
	 */
	public function getEncodingOfHeaders(){
		return $this->getHeaderEncoding();
	}

	/**
	 * Return the encoding of mail headers
	 *
	 * Either Mime::ENCODING_QUOTEDPRINTABLE or Mime::ENCODING_BASE64
	 *
	 * @return string
	 */
	public function getHeaderEncoding(){
		return $this->_headerEncoding;
	}

	/**
	 * Set the encoding of mail headers
	 *
	 * @deprecated Use {@link setHeaderEncoding()} instead.
	 * @param  string $encoding
	 * @return we_mail_znd
	 */
	public function setEncodingOfHeaders($encoding){
		return $this->setHeaderEncoding($encoding);
	}

	/**
	 * Set the encoding of mail headers
	 *
	 * @param  string $encoding Mime::ENCODING_QUOTEDPRINTABLE or Mime::ENCODING_BASE64
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setHeaderEncoding($encoding){
		$allowed = [we_mail_mime::ENCODING_BASE64,
			we_mail_mime::ENCODING_QUOTEDPRINTABLE
		];
		if(!in_array($encoding, $allowed)){
			/**
			 * @see Mail_Exception
			 */
			throw new we_mail_exception('Invalid encoding "' . $encoding . '"');
		}
		$this->_headerEncoding = $encoding;

		return $this;
	}

	/**
	 * Sets the text body for the message.
	 *
	 * @param  string $txt
	 * @param  string $charset
	 * @param  string $encoding
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setBodyText($txt, $charset = null, $encoding = we_mail_mime::ENCODING_QUOTEDPRINTABLE){
		if($charset === null){
			$charset = $this->_charset;
		}

		$mp = new we_mail_mimePart($txt);
		$mp->encoding = $encoding;
		$mp->type = we_mail_mime::TYPE_TEXT;
		$mp->disposition = we_mail_mime::DISPOSITION_INLINE;
		$mp->charset = $charset;

		$this->_bodyText = $mp;

		return $this;
	}

	/**
	 * Return text body Mime_Part or string
	 *
	 * @param  bool textOnly Whether to return just the body text content or the MIME part; defaults to false, the MIME part
	 * @return false|we_mail_mimePart|string
	 */
	public function getBodyText($textOnly = false){
		if($textOnly && $this->_bodyText){
			$body = $this->_bodyText;
			return $body->getContent();
		}

		return $this->_bodyText;
	}

	/**
	 * Sets the HTML body for the message
	 *
	 * @param  string    $html
	 * @param  string    $charset
	 * @param  string    $encoding
	 * @return we_mail_znd Provides fluent interface
	 */
	protected function setBodyHtml(we_mail_mimePart $html){
		$this->_bodyHtml = $html;
	}

	/**
	 * Return Mime_Part representing body HTML
	 *
	 * @param  bool $htmlOnly Whether to return the body HTML only, or the MIME part; defaults to false, the MIME part
	 * @return false|we_mail_mimePart|string
	 */
	public function getBodyHtml($htmlOnly = false){
		if($htmlOnly && $this->_bodyHtml){
			$body = $this->_bodyHtml;
			return $body->getContent();
		}

		return $this->_bodyHtml;
	}

	/**
	 * Adds an existing attachment to the mail message
	 *
	 * @param  we_mail_mimePart $attachment
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addAttachment(we_mail_mimePart $attachment){
		$this->addPart($attachment);
		$this->hasAttachments = true;

		return $this;
	}

	/**
	 * Creates a Mime_Part attachment
	 *
	 * Attachment is automatically added to the mail object after creation. The
	 * attachment object is returned to allow for further manipulation.
	 *
	 * @param  string         $body
	 * @param  string         $mimeType
	 * @param  string         $disposition
	 * @param  string         $encoding
	 * @param  string         $filename OPTIONAL A filename for the attachment
	 * @return we_mail_mimePart Newly created Mime_Part object (to allow
	 * advanced settings)
	 */
	public function createAttachment($body, $mimeType = we_mail_mime::TYPE_OCTETSTREAM, $disposition = we_mail_mime::DISPOSITION_ATTACHMENT, $encoding = we_mail_mime::ENCODING_BASE64, $filename = null){

		$mp = new we_mail_mimePart($body);
		$mp->encoding = $encoding;
		$mp->type = $mimeType;
		$mp->disposition = $disposition;
		$mp->filename = $filename;

		$this->addAttachment($mp);

		return $mp;
	}

	/**
	 * Return a count of message parts
	 *
	 * @return integer
	 */
	public function getPartCount(){
		return count($this->_parts);
	}

	/**
	 * Encode header fields
	 *
	 * Encodes header content according to RFC1522 if it contains non-printable
	 * characters.
	 *
	 * @param  string $value
	 * @return string
	 */
	protected function _encodeHeader($value){
		if(we_mail_mime::isPrintable($value) === false){
			return
				($this->getHeaderEncoding() === we_mail_mime::ENCODING_QUOTEDPRINTABLE ?
				we_mail_mime::encodeQuotedPrintableHeader($value, $this->getCharset(), we_mail_mime::LINELENGTH, we_mail_mime::LINEEND) :
				we_mail_mime::encodeBase64Header($value, $this->getCharset(), we_mail_mime::LINELENGTH, we_mail_mime::LINEEND)
				);
		}

		return $value;
	}

	/**
	 * Add a header to the message
	 *
	 * Adds a header to this message. If append is true and the header already
	 * exists, raises a flag indicating that the header should be appended.
	 *
	 * @param string  $headerName
	 * @param string  $value
	 * @param bool $append
	 */
	protected function _storeHeader($headerName, $value, $append = false){
		if(isset($this->_headers[$headerName])){
			$this->_headers[$headerName][] = $value;
		} else {
			$this->_headers[$headerName] = [$value];
		}

		if($append){
			$this->_headers[$headerName]['append'] = true;
		}
	}

	/**
	 * Clear header from the message
	 *
	 * @param string $headerName
	 * @deprecated use public method directly
	 */
	protected function _clearHeader($headerName){
		$this->clearHeader($headerName);
	}

	/**
	 * Helper function for adding a recipient and the corresponding header
	 *
	 * @param string $headerName
	 * @param string $email
	 * @param string $name
	 */
	protected function _addRecipientAndHeader($headerName, $email, $name){
		$email = $this->_filterEmail($email);
		// prevent duplicates
		$this->_recipients[$email] = 1;
		$this->_storeHeader($headerName, $this->_formatAddress($email, $this->_filterName($name)), true);
	}

	/**
	 * Adds To-header and recipient, $email can be an array, or a single string address
	 *
	 * @param  string|array $email
	 * @param  string $name
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addTo($email, $name = ''){
		if(!is_array($email)){
			$email = [$name => $email];
		}

		foreach($email as $n => $recipient){
			$this->_addRecipientAndHeader('To', $recipient, is_int($n) ? '' : $n);
			$this->_to[] = $recipient;
		}

		return $this;
	}

	/**
	 * Adds Cc-header and recipient, $email can be an array, or a single string address
	 *
	 * @param  string|array    $email
	 * @param  string    $name
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addCc($email, $name = ''){
		if(!is_array($email)){
			$email = [$name => $email];
		}

		foreach($email as $n => $recipient){
			$this->_addRecipientAndHeader('Cc', $recipient, is_int($n) ? '' : $n);
		}

		return $this;
	}

	/**
	 * Adds Bcc recipient, $email can be an array, or a single string address
	 *
	 * @param  string|array    $email
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addBcc($email, $name = ''){
		if(!is_array($email)){
			$email = [$name => $email];
		}
		foreach($email as $n => $recipient){
			$this->_addRecipientAndHeader('Bcc', $recipient, is_int($n) ? '' : $n);
		}

		return $this;
	}

	/**
	 * Return list of recipient email addresses
	 *
	 * @return array (of strings)
	 */
	public function getRecipients(){
		return array_keys($this->_recipients);
	}

	/**
	 * Clear header from the message
	 *
	 * @param string $headerName
	 * @return we_mail_znd Provides fluent inter
	 */
	public function clearHeader($headerName){
		if(isset($this->_headers[$headerName])){
			unset($this->_headers[$headerName]);
		}
		return $this;
	}

	/**
	 * Clears list of recipient email addresses
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearRecipients(){
		$this->_recipients = [];
		$this->_to = [];

		$this->clearHeader('To');
		$this->clearHeader('Cc');
		$this->clearHeader('Bcc');

		return $this;
	}

	/**
	 * Sets From-header and sender of the message
	 *
	 * @param  string    $email
	 * @param  string    $name
	 * @return we_mail_znd Provides fluent interface
	 * @throws we_mail_exception if called subsequent times
	 */
	public function setFrom($email, $name = null){
		if(null !== $this->_from){
			/**
			 * @see Mail_Exception
			 */
			throw new we_mail_exception('From Header set twice');
		}

		$email = $this->_filterEmail($email);
		$name = $this->_filterName($name);
		$this->_from = $email;
		$this->_storeHeader('From', $this->_formatAddress($email, $name), true);

		return $this;
	}

	/**
	 * Set Reply-To Header
	 *
	 * @param string $email
	 * @param string $name
	 * @return we_mail_znd
	 * @throws we_mail_exception if called more than one time
	 */
	public function setReplyTo($email, $name = null){
		if(null !== $this->_replyTo){
			throw new we_mail_exception('Reply-To Header set twice');
		}

		$this->_replyTo = $this->_filterEmail($email);
		$this->_storeHeader('Reply-To', $this->_formatAddress($this->_replyTo, $this->_filterName($name)), true);

		return $this;
	}

	/**
	 * Returns the sender of the mail
	 *
	 * @return string
	 */
	public function getFrom(){
		return $this->_from;
	}

	/**
	 * Returns the current Reply-To address of the message
	 *
	 * @return string|null Reply-To address, null when not set
	 */
	public function getReplyTo(){
		return $this->_replyTo;
	}

	/**
	 * Clears the sender from the mail
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearFrom(){
		$this->_from = null;
		$this->clearHeader('From');

		return $this;
	}

	/**
	 * Clears the current Reply-To address from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearReplyTo(){
		$this->_replyTo = null;
		$this->clearHeader('Reply-To');

		return $this;
	}

	/**
	 * Sets Default From-email and name of the message
	 *
	 * @param  string               $email
	 * @param  string    Optional   $name
	 * @return void
	 */
	public static function setDefaultFrom($email, $name = null){
		self::$defaultFrom = ['email' => $email, 'name' => $name];
	}

	/**
	 * Returns the default sender of the mail
	 *
	 * @return null|array   Null if none was set.
	 */
	public static function getDefaultFrom(){
		return self::$defaultFrom;
	}

	/**
	 * Clears the default sender from the mail
	 *
	 * @return void
	 */
	public static function clearDefaultFrom(){
		self::$defaultFrom = null;
	}

	/**
	 * Sets From-name and -email based on the defaults
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setFromToDefaultFrom(){
		$from = self::getDefaultFrom();
		if($from === null){
			throw new we_mail_exception('No default From Address set to use');
		}

		$this->setFrom($from['email'], $from['name']);

		return $this;
	}

	/**
	 * Sets Default ReplyTo-address and -name of the message
	 *
	 * @param  string               $email
	 * @param  string    Optional   $name
	 * @return void
	 */
	public static function setDefaultReplyTo($email, $name = null){
		self::$defaultReplyTo = ['email' => $email, 'name' => $name];
	}

	/**
	 * Returns the default Reply-To Address and Name of the mail
	 *
	 * @return null|array   Null if none was set.
	 */
	public static function getDefaultReplyTo(){
		return self::$defaultReplyTo;
	}

	/**
	 * Clears the default ReplyTo-address and -name from the mail
	 *
	 * @return void
	 */
	public static function clearDefaultReplyTo(){
		self::$defaultReplyTo = null;
	}

	/**
	 * Sets ReplyTo-name and -email based on the defaults
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setReplyToFromDefault(){
		$replyTo = self::getDefaultReplyTo();
		if($replyTo === null){
			throw new we_mail_exception('No default Reply-To Address set to use');
		}

		$this->setReplyTo($replyTo['email'], $replyTo['name']);

		return $this;
	}

	/**
	 * Sets the Return-Path header of the message
	 *
	 * @param  string    $email
	 * @return we_mail_znd Provides fluent interface
	 * @throws we_mail_exception if set multiple times
	 */
	public function setReturnPath($email){
		if($this->_returnPath === null){
			$email = $this->_filterEmail($email);
			$this->_returnPath = $email;
			$this->_storeHeader('Return-Path', $email, false);
		} else {
			/**
			 * @see Mail_Exception
			 */
			throw new we_mail_exception('Return-Path Header set twice');
		}
		return $this;
	}

	/**
	 * Returns the current Return-Path address of the message
	 *
	 * If no Return-Path header is set, returns the value of {@link $_from}.
	 *
	 * @return string
	 */
	public function getReturnPath(){
		if(null !== $this->_returnPath){
			return $this->_returnPath;
		}

		return $this->_from;
	}

	/**
	 * Clears the current Return-Path address from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearReturnPath(){
		$this->_returnPath = null;
		$this->clearHeader('Return-Path');

		return $this;
	}

	/**
	 * Sets the subject of the message
	 *
	 * @param   string    $subject
	 * @return  we_mail_znd Provides fluent interface
	 * @throws  we_mail_exception
	 */
	public function setSubject($subject){
		if($this->_subject === null){
			$subject = $this->_filterOther($subject);
			$this->_subject = $this->_encodeHeader($subject);
			$this->_storeHeader('Subject', $this->_subject);
		} else {
			/**
			 * @see Mail_Exception
			 */
			throw new we_mail_exception('Subject set twice');
		}
		return $this;
	}

	/**
	 * Returns the encoded subject of the message
	 *
	 * @return string
	 */
	public function getSubject(){
		return $this->_subject;
	}

	/**
	 * Clears the encoded subject from the message
	 *
	 * @return  we_mail_znd Provides fluent interface
	 */
	public function clearSubject(){
		$this->_subject = null;
		$this->clearHeader('Subject');

		return $this;
	}

	/**
	 * Sets Date-header
	 *
	 * @param  timestamp|string|Date $date
	 * @return we_mail_znd Provides fluent interface
	 * @throws we_mail_exception if called subsequent times or wrong date format.
	 */
	public function setDate($date = null){
		if($this->_date === null){
			if($date === null){
				$date = date('r');
			} else if(is_int($date)){
				$date = date('r', $date);
			} else if(is_string($date)){
				$date = strtotime($date);
				if($date === false || $date < 0){
					/**
					 * @see Mail_Exception
					 */
					throw new we_mail_exception('String representations of Date Header must be strtotime()-compatible');
				}
				$date = date('r', $date);
			} else {
				/**
				 * @see Mail_Exception
				 */
				throw new we_mail_exception(__METHOD__ . ' only accepts UNIX timestamps and strtotime()-compatible strings');
			}
			$this->_date = $date;
			$this->_storeHeader('Date', $date);
		} else {
			throw new we_mail_exception('Date Header set twice');
		}
		return $this;
	}

	/**
	 * Returns the formatted date of the message
	 *
	 * @return string
	 */
	public function getDate(){
		return $this->_date;
	}

	/**
	 * Clears the formatted date from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearDate(){
		$this->_date = null;
		$this->clearHeader('Date');

		return $this;
	}

	/**
	 * Sets the Message-ID of the message
	 *
	 * @param   boolean|string  $id
	 * true  :Auto
	 * false :No set
	 * null  :No set
	 * string:Sets given string (Angle brackets is not necessary)
	 * @return  we_mail_znd Provides fluent interface
	 * @throws  we_mail_exception
	 */
	public function setMessageId($id = true){
		if($id === null || $id === false){
			return $this;
		} elseif($id === true){
			$id = $this->createMessageId();
		}

		if($this->_messageId === null){
			$id = $this->_filterOther($id);
			$this->_messageId = $id;
			$this->_storeHeader('Message-Id', '<' . $this->_messageId . '>');
		} else {
			throw new we_mail_exception('Message-ID set twice');
		}

		return $this;
	}

	/**
	 * Returns the Message-ID of the message
	 *
	 * @return string
	 */
	public function getMessageId(){
		return $this->_messageId;
	}

	/**
	 * Clears the Message-ID from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearMessageId(){
		$this->_messageId = null;
		$this->clearHeader('Message-Id');

		return $this;
	}

	/**
	 * Creates the Message-ID
	 *
	 * @return string
	 */
	public function createMessageId(){
		$time = time();

		$user = ($this->_from !== null ?
			$this->_from :
			(isset($_SERVER['REMOTE_ADDR']) ?
			$_SERVER['REMOTE_ADDR'] :
			getmypid())
			);

		$rand = mt_rand();

		$recipient = ($this->_recipients !== [] ?
			array_rand($this->_recipients) :
			'unknown');

		$hostName = (isset($_SERVER['SERVER_NAME']) ?
			$_SERVER['SERVER_NAME'] :
			php_uname('n'));


		return sha1($time . $user . $rand . $recipient) . '@' . $hostName;
	}

	/**
	 * Add a custom header to the message
	 *
	 * @param  string              $name
	 * @param  string              $value
	 * @param  boolean             $append
	 * @return we_mail_znd           Provides fluent interface
	 * @throws we_mail_exception on attempts to create standard headers
	 */
	public function addHeader($name, $value, $append = false){
		if(in_array(strtolower($name), self::PROHIBITTED_MAIL_HEADERS)){
			throw new we_mail_exception('Cannot set standard header from addHeader()');
		}

		$this->_storeHeader($name, $this->_encodeHeader($this->_filterOther($value)), $append);

		return $this;
	}

	/**
	 * Return mail headers
	 *
	 * @return array
	 */
	public function getHeaders(){
		return $this->_headers;
	}

	/**
	 * Sends this email using the given transport or a previously
	 * set DefaultTransport or the internal mail function if no
	 * default transport had been set.
	 *
	 * @param  we_mail_TransportAbstract $transport
	 * @return we_mail_znd                    Provides fluent interface
	 */
	public function send(){
		$transport = (!self::$defaultTransport instanceof we_mail_TransportAbstract ?
			new we_mail_TransportSendmail() :
			self::$defaultTransport);

		if($this->_date === null){
			$this->setDate();
		}

		if(null === $this->_from && null !== self::getDefaultFrom()){
			$this->setFromToDefaultFrom();
		}

		if(null === $this->_replyTo && null !== self::getDefaultReplyTo()){
			$this->setReplyToFromDefault();
		}

		$transport->send($this);

		return $this;
	}

	/**
	 * Filter of email data
	 *
	 * @param string $email
	 * @return string
	 */
	protected function _filterEmail($email){
		return strtr($email, [
			"\r" => '',
			"\n" => '',
			"\t" => '',
			'"' => '',
			',' => '',
			'<' => '',
			'>' => '',
		]);
	}

	/**
	 * Filter of name data
	 *
	 * @param string $name
	 * @return string
	 */
	protected function _filterName($name){
		return trim(strtr($name, [
			"\r" => '',
			"\n" => '',
			"\t" => '',
			'"' => "'",
			'<' => '[',
			'>' => ']',
		]));
	}

	/**
	 * Filter of other data
	 *
	 * @param string $data
	 * @return string
	 */
	protected function _filterOther($data){
		return strtr($data, [
			"\r" => '',
			"\n" => '',
			"\t" => '',
		]);
	}

	/**
	 * Formats e-mail address
	 *
	 * @param string $email
	 * @param string $name
	 * @return string
	 */
	protected function _formatAddress($email, $name){
		if($name === '' || $name === null || $name === $email){
			return $email;
		}
		$encodedName = $this->_encodeHeader($name);
		$format = ($encodedName === $name && strcspn($name, '()<>[]:;@\\,.') != strlen($name) ?
			'"%s" <%s>' :
			'%s <%s>');

		return sprintf($format, $encodedName, $email);
	}

}
